; ######################
; # BETA-REDUCTION LISP LIBRARY #
; ######################
; Author: Dipl.Inf. (FH) Thomas Wllert, Dipl.Inf. (FH) Stefan Hertel
; EMail: thomas@woellert.net, sh_ccbiker@gmx.de
;
; Munich University of Applied Sciences, FB07 Computer Science
; Semantik von Programmiersprachen - Prof.Dr.Schiedermeier - Master 1.Semester

;################################################
; Exercise 4.2.2
; Checks if the given expression is a beta-redex => ((LAMBDA (x) (E)) M). Needs a parsed
; lambda calculus expression.
; i.e. ((LAMBDA (X) (X Y)) Z) => [APP [LAMBDA X [APP [VARREF X] [VARREF Y]]] [VARREF Z]]
; more general: ((LAMBDA (anyVar) (anyExpr)) anyExpr) => [APP [LAMBDA anyVar [anyExpr]] [anyExpr]]
;
; USAGE: (beta-redex? <parsed-lambda-calculus-expression>)
; EXAMPLE: (beta-redex? (parse '((LAMBDA (x) (y x)) z)));
(def beta-redex?
	(lambda (exp)
		(if (null? (vector? exp))
			(error 'ERR:beta-redex?:given-expression-must-be-parsed-before-calling-beta-redex?)
			(variant-case exp
				'(
					(APP (rator rand)
						(variant-case rator
							'(
								(LAMBDA (formal body)
									t
								)
								(else nil)
							)
						)
					)
					(else nil)
				)
			)
		)
	)
);

;################################################
; Exercise 4.2.3
; Performs the substitution e[m/x] (x is properly replaced by m) for the given expressions
; e and m and the given variable x. Needs parsed lambda calculus expressions
;
; USAGE: (substitute <parsed-lambda-calculus-expression> <parsed-lambda-calculus-expression> 'x)
; EXAMPLE: (unparse (substitute (parse '(a b)) (parse 'c) 'b));
(def substitute
	(lambda (exprE exprM varX)
		(if (null? (vector? exprE))
			(error 'ERR:substitute:given-first-expression-must-be-parsed-before-calling-substitute)
			(if (null? (vector? exprM))
				(error 'ERR:substitute:given-second-expression-must-be-parsed-before-calling-substitute)
				(variant-case exprE
					`(
						(VARREF (var)
							(if (eq? var ',varX)
								',exprM
								',exprE
							)
						)
						(LAMBDA (formal body)
							(cond
								((eq? formal ',varX) 
									',exprE
								)
								((null? (FREE? body ',varX)) 
									',exprE
								)
								((null? (eq? formal ',varX)) 
									(if (null? (FREE? ',exprM formal)) 
										(make-LAMBDA (list formal (substitute body ',exprM ',varX)))
										(if (FREE? body ',varX)
											(substitute (alpha-convert-ast ',exprE (symbol)) ',exprM ',varX)
											(error 'ERR:substitute:undecided:y-isNotEqual-x-and-y-isFreeInM-but-x-isNotFreeInEDash)
										)
									)
								)
								(t (error 'ERR:substitute:should-never-hit-because-X-either-equals-Y-or-not))
							)
						)
						(APP (rator rand)
							(make-APP (list (substitute rator ',exprM ',varX) (substitute rand ',exprM ',varX)))
						)
						(else (error 'substitute:invalid-abstract-syntax))
					)
				)
			)
		)
	)
);

;################################################
; Exercise 4.2.4
; Performs the beta-reduction rule for the given beta-redex expression
; Rule: ((lambda (x) E) M) = E[M/x]
; Needs a parsed lambda calculus beta-redex expression
;
; USAGE: (beta-reduce <parsed-lambda-calculus-beta-redex-expression>)
; EXAMPLE: (unparse (beta-reduce (parse '((LAMBDA (x) (y x)) z))));
(def beta-reduce
	(lambda (exp)
		(if (null? (vector? exp))
			(error 'ERR:beta-reduce:given-expression-must-be-parsed-before-calling-beta-reduce)
			(if (null? (beta-redex? exp))
				(error 'ERR:beta-reduce:given-expression-is-no-beta-redex)
				(variant-case exp
					'(
						(APP (rator rand)
							(variant-case rator
								`(
									(LAMBDA (formal body)
										(substitute body ',rand formal)
									)
									(else (error 'ERR:beta-reduce:no-lambda-found-in-given-expression))
								)
							)
						)
						(else (error 'ERR:beta-reduce:no-application-found-in-given-expression))
					)
				)
			)
		)
	)
);
