; #####################
; # ETA-REDUCTION LISP LIBRARY #
; #####################
; Author: Dipl.Inf. (FH) Thomas Wllert, Dipl.Inf. (FH) Stefan Hertel
; EMail: thomas@woellert.net, sh_ccbiker@gmx.de
;
; Munich University of Applied Sciences, FB07 Computer Science
; Semantik von Programmiersprachen - Prof.Dr.Schiedermeier - Master 1.Semester

;################################################

; Exercise 4.2.5
; Checks if the given expression is a eta-redex => (LAMBDA (x) (E x)) => E. Needs a parsed
; lambda calculus expression.
;
; USAGE: (eta-redex? <parsed-lambda-calculus-expression>)
; EXAMPLE: (eta-redex? (parse '(LAMBDA (x) (y x)));
(def eta-redex?
	(lambda (exp)
		(if (null? (vector? exp))
			(error 'ERR:beta-redex?:given-expression-must-be-parsed-before-calling-beta-redex?)
			(variant-case exp
				'(
					(LAMBDA (formal body)
						(let ((form formal))
						(variant-case body
							`(
								(APP (rator rand)
									(let ((freelist (FREE_VARS rator)))
										(if (inlist? freelist ',form)
											nil
											(variant-case rand
												'(
													(VARREF (var)
														(= var ',form)
													)
													(else nil)
												)
											)
										)
									)
								)
								(else nil)
							)
						)
						)
					)
					(else nil)
				)
			)
		)
	)
);

;################################################
; Exercise 4.2.6
; Performs the eta-reduction rule for the given eta-redex expression
; Rule: (lambda (x) E x) = E
; Needs a parsed lambda calculus beta-redex expression
;
; USAGE: (eta-reduce <parsed-lambda-calculus-beta-redex-expression>)
; EXAMPLE: (unparse (eta-reduce (parse '(LAMBDA (x) (y x))));
(def eta-reduce
	(lambda (exp)
		(if (null? (vector? exp))
			(error 'ERR:eta-reduce:given-expression-must-be-parsed-before-calling-eta-reduce)
			(if (null? (eta-redex? exp))
				(error 'ERR:eta-reduce:given-expression-is-no-eta-redex)
				(variant-case exp
					'(
						(LAMBDA (formal body)
							(variant-case body
								`(
									(APP (rator rand)
										rator
									)
									(else exp)
								)
							)
						)
						(else exp)
					)
				)
			)
		)
	)
);

;################################################
; Exercise 4.2.7
; Performs the eta-expand rule for a given expresion 
; Rule: E = (lambda (x) E x)
; Needs a parsed lambda calculus expression
;
; USAGE: (eta-expand <parsed-lambda-calculus-expression>)
; EXAMPLE: (unparse (eta-expand (parse '(LAMBDA (y) y))));
(def eta-expand
	(lambda (exp)
		(let ((newsym (symbol)))
			(make-LAMBDA (list newsym (make-APP (list exp (make-VARREF (list newsym))))))
		)
	)
);
